@@ -44,6 +44,7 @@ let s_path ctx path p =
 		| "Float" -> "Number"
 		| "Dynamic" -> "Object"
 		| "Bool" -> "Boolean"
+		| "Enum" -> "Class"
 		| _ -> name)
 	| (["flash"],"FlashXml__") ->
 		"Xml"
@@ -62,6 +63,7 @@ let s_path ctx path p =
 let s_ident n =
 	match n with
 	| "is" -> "_is"
+	| "as" -> "_as"
 	| "int" -> "_int"
 	| "getTimer" -> "_getTimer"
 	| "typeof" -> "_typeof"
@@ -133,18 +135,6 @@ let print ctx = Printf.kprintf (fun s -> Buffer.add_string ctx.buf s)
 
 let unsupported = Typer.error "This expression cannot be generated to AS3"
 
-let rec follow_not_stat t =
-	match t with
-	| TMono r ->
-		(match !r with
-		| Some t -> follow_not_stat t
-		| _ -> t)
-	| TLazy f ->
-		follow_not_stat (!f())
-	| TType (t,tl) when t.t_static = None ->
-		follow_not_stat (apply_params t.t_types tl t.t_type)
-	| _ -> t
-
 let newline ctx =
 	match Buffer.nth ctx.buf (Buffer.length ctx.buf - 1) with
 	| '}' | '{' | ':' -> print ctx "\n%s" ctx.tabs
@@ -178,6 +168,7 @@ let rec type_str ctx t p =
 		if e.e_extern then (match e.e_path with
 			| [], "Void" -> "void"
 			| [], "Bool" -> "Boolean"
+			| "flash" :: _ , _ -> "String"
 			| _ -> "Object"
 		) else
 			s_path ctx e.e_path p
@@ -333,9 +324,8 @@ and gen_value_op ctx e =
 		gen_value ctx e
 
 and gen_field_access ctx t s =
-	match follow_not_stat t with
-	| TType ({ t_static = Some c },_) | TInst (c,_) ->
-		(match fst c.cl_path, snd c.cl_path, s with
+	let field c =
+		match fst c.cl_path, snd c.cl_path, s with
 		| [], "Math", "NaN"
 		| [], "Math", "NEGATIVE_INFINITY"
 		| [], "Math", "POSITIVE_INFINITY"
@@ -349,7 +339,14 @@ and gen_field_access ctx t s =
 		->
 			print ctx "[\"%s\"]" s
 		| _ ->
-			print ctx ".%s" (s_ident s));
+			print ctx ".%s" (s_ident s)
+	in
+	match follow t with
+	| TInst (c,_) -> field c
+	| TAnon a ->
+		(match !(a.a_status) with
+		| Statics c -> field c
+		| _ -> print ctx ".%s" (s_ident s))
 	| _ ->
 		print ctx ".%s" (s_ident s)
 
@@ -397,6 +394,10 @@ and gen_expr ctx e =
 		(match eo with
 		| None ->
 			spr ctx "return"
+		| Some e when (match follow e.etype with TEnum({ e_path = [],"Void" },[]) -> true | _ -> false) ->
+			gen_value ctx e;
+			newline ctx;
+			spr ctx "return"
 		| Some e ->
 			spr ctx "return ";
 			gen_value ctx e);
@@ -522,11 +523,11 @@ and gen_expr ctx e =
 		print ctx "var %s : enum = " tmp;
 		gen_value ctx e;
 		newline ctx;
-		print ctx "switch( %s.tag ) {" tmp;
+		print ctx "switch( %s.index ) {" tmp;
 		newline ctx;
 		List.iter (fun (cl,params,e) ->
 			List.iter (fun c ->
-				print ctx "case \"%s\":" c;
+				print ctx "case %d:" c;
 				newline ctx;
 			) cl;
 			let b = save_locals ctx in
@@ -714,7 +715,7 @@ let generate_field ctx static f =
 	ctx.in_static <- static;
 	ctx.locals <- PMap.empty;
 	ctx.inv_locals <- PMap.empty;
-	let public = f.cf_public || Hashtbl.mem ctx.get_sets (f.cf_name,static) || (f.cf_name = "main" && static) in
+	let public = f.cf_public || Hashtbl.mem ctx.get_sets (f.cf_name,static) || (f.cf_name = "main" && static) || f.cf_name = "__resolve" in
 	let rights = (if static then "static " else "") ^ (if public then "public" else "protected") in
 	let p = ctx.curclass.cl_pos in
 	match f.cf_expr with
@@ -746,7 +747,28 @@ let generate_field ctx static f =
 				) args;
 				print ctx ") : %s " (type_str ctx r p);
 			| _ -> ()
-		else begin
+		else
+		if (match f.cf_get with MethodAccess m -> true | _ -> match f.cf_set with MethodAccess m -> true | _ -> false) then begin
+			let t = type_str ctx f.cf_type p in
+			let id = s_ident f.cf_name in
+			(match f.cf_get with
+			| NormalAccess ->
+				print ctx "%s function get %s() : %s { return $%s; }" rights id t id;
+				newline ctx
+			| MethodAccess m ->
+				print ctx "%s function get %s() : %s { return %s(); }" rights id t m;
+				newline ctx
+			| _ -> ());
+			(match f.cf_set with
+			| NormalAccess ->
+				print ctx "%s function set %s( __v : %s ) : void { $%s = __v; }" rights id t id;
+				newline ctx
+			| MethodAccess m ->
+				print ctx "%s function set %s( __v : %s ) : void { %s(__v); }" rights id t m;
+				newline ctx
+			| _ -> ());
+			print ctx "protected var $%s : %s" (s_ident f.cf_name) (type_str ctx f.cf_type p);
+		end else begin
 			print ctx "%s var %s : %s" rights (s_ident f.cf_name) (type_str ctx f.cf_type p);
 			match f.cf_expr with
 			| None -> ()
@@ -775,7 +797,7 @@ let generate_class ctx c =
 	(match c.cl_implements with
 	| [] -> ()
 	| l ->
-		spr ctx "implements ";
+		spr ctx (if c.cl_interface then "extends " else "implements ");
 		concat ctx ", " (fun (i,_) -> print ctx "%s" (s_path ctx i.cl_path c.cl_pos)) l);
 	spr ctx "{";
 	let cl = open_block ctx in
@@ -787,18 +809,8 @@ let generate_class ctx c =
 			cf_public = true;
 			cf_set = F9MethodAccess;
 		} in
-		let fd = (match f.cf_expr with Some { eexpr = TFunction f } -> f | _ -> assert false) in
 		ctx.constructor_block <- true;
 		generate_field ctx false f;
-		newline ctx;
-		print ctx "public static function __construct__(args:Array) : %s {" (snd c.cl_path);
-		newline ctx;
-		print ctx "\treturn new %s(" (snd c.cl_path);
-		let n = ref 0 in
-		concat ctx "," (fun _ -> print ctx "args[%d]" !n; incr n) fd.tf_args;
-		spr ctx ")";
-		newline ctx;
-		spr ctx "}";
 	);
 	List.iter (generate_field ctx false) c.cl_ordered_fields;
 	List.iter (generate_field ctx true) c.cl_ordered_statics;
@@ -845,7 +857,7 @@ let generate_enum ctx e =
 	newline ctx;
 	print ctx "public static const __isenum : Boolean = true";
 	newline ctx;
-	print ctx "public function %s( t : String, p : Array = null ) : void { this.tag = t; this.params = p; }" ename;
+	print ctx "public function %s( t : String, index : int, p : Array = null ) : void { this.tag = t; this.index = index; this.params = p; }" ename;
 	PMap.iter (fun _ c ->
 		newline ctx;
 		match c.ef_type with
@@ -856,11 +868,11 @@ let generate_enum ctx e =
 				if o then spr ctx " = null";
 			) args;
 			print ctx ") : %s {" ename;
-			print ctx " return new %s(\"%s\",[" ename c.ef_name;
+			print ctx " return new %s(\"%s\",%d,[" ename c.ef_name c.ef_index;
 			concat ctx "," (fun (a,_,_) -> spr ctx a) args;
 			print ctx "]); }";
 		| _ ->
-			print ctx "public static var %s : %s = new %s(\"%s\")" c.ef_name ename ename c.ef_name;
+			print ctx "public static var %s : %s = new %s(\"%s\",%d)" c.ef_name ename ename c.ef_name c.ef_index;
 	) e.e_constrs;
 	cl();
 	newline ctx;
@@ -877,6 +889,8 @@ let generate_base_enum ctx =
 	newline ctx;
 	spr ctx "public var tag : String";
 	newline ctx;
+	spr ctx "public var index : int";
+	newline ctx;
 	spr ctx "public var params : Array";
 	cl();
 	newline ctx;
@@ -932,3 +946,233 @@ let generate dir types =
 		ctx.inits <- List.rev !inits;
 		generate_class ctx c;
 		close ctx
+
+(* ----------------------------------------------------------------------------------------
+
+	HX generation
+
+   ---------------------------------------------------------------------------------------- *)
+open As3
+
+let s_type_path = Ast.s_type_path
+
+let ident ctx p =
+	As3code.iget ctx.as3_idents p
+
+let package ctx idx =
+	match As3code.iget ctx.as3_namespaces idx with
+	| A3NPrivate (Some id)
+	| A3NPublic (Some id)
+	| A3NInternal (Some id)
+	| A3NProtected id
+	| A3NExplicit id
+	| A3NStaticProtected (Some id) ->
+		let pack = ident ctx id in
+		ExtString.String.nsplit pack "."
+	| A3NNamespace id ->
+		["/* namespace " ^ ident ctx id ^ "*/"]
+	| A3NPrivate None | A3NPublic None | A3NInternal None | A3NStaticProtected None ->
+		[]
+
+let rec real_type_path ctx p =
+	let rec loop = function
+		| A3MName (id,pack) ->
+			let name = ident ctx id in
+			let pack = package ctx pack in
+			pack , name
+		| A3MMultiName (Some id,pack) ->
+			let name = ident ctx id in
+			let pack = package ctx (List.hd (As3code.iget ctx.as3_nsets pack)) in
+			pack , name
+		| A3MMultiName (None,_) ->
+			[] , "$MultiName"
+		| A3MMultiNameLate _ ->
+			[] , "$MultiNameLate"
+		| A3MRuntimeName _ ->
+			[] , "$RuntimeName"
+		| A3MRuntimeNameLate ->
+			[] , "$RuntimeNameLate"
+		| A3MAttrib n ->
+			let path, name = loop n in
+			"$Attrib" :: path, name
+	in
+	loop (As3code.iget ctx.as3_names p)
+
+let type_path ctx p =
+	match real_type_path ctx p with
+	| [] , "Object" -> [] , "Dynamic"
+	| [] , "Boolean" -> [] , "Bool"
+	| [] , "int" -> [] , "Int"
+	| [] , "uint" -> [] , "UInt"
+	| [] , "Number" -> [] , "Float"
+	| [] , "Array" -> [] , "Array<Dynamic>"
+	| [] , "void" -> [] , "Void"
+	| [] , "Function" -> [] , "Dynamic"
+	| [] , "Class" -> [] , "Class<Dynamic>"
+	| path -> path
+
+let ident_rights ctx id =
+	match As3code.iget ctx.as3_names id with
+	| A3MName (id,r) ->
+		let name = ident ctx id in
+		(match As3code.iget ctx.as3_namespaces r with
+		| A3NNamespace i when As3code.iget ctx.as3_idents i = "http://www.adobe.com/2006/flex/mx/internal" -> false, "$" ^ name
+		| A3NPublic _ | A3NNamespace _ -> false , name
+		| _ -> true , name)
+	| _ -> false, "???"
+
+let rec create_dir acc = function
+	| [] -> ()
+	| d :: l ->
+		let path = acc ^ "/" ^ d in
+		(try Unix.mkdir path 0o777 with _ -> ());
+		create_dir path l
+
+let value_type = function
+	| A3VNone
+	| A3VNull -> "Dynamic"
+	| A3VBool _ -> "Bool"
+	| A3VString _ -> "String"
+	| A3VInt _ -> "Int"
+	| A3VUInt _ -> "UInt"
+	| A3VFloat _ -> "Float"
+	| A3VNamespace _ -> "$Namespace"
+
+let type_val ctx t v =
+	match t with
+	| None ->
+		(match v with
+		| None -> "Dynamic"
+		| Some v -> value_type v)
+	| Some t ->
+		s_type_path (type_path ctx t)
+
+let has_getset ml f m =
+	List.exists (fun f2 ->
+		match f2.f3_kind with
+		| A3FMethod m2 when f.f3_name = f2.f3_name ->
+			(match m.m3_kind , m2.m3_kind with
+			| MK3Getter , MK3Setter | MK3Setter , MK3Getter -> true
+			| _ -> false)
+		| _ -> false
+	) ml
+
+let gen_method ctx ch name mt =
+	let m = As3code.iget ctx.as3_method_types (As3parse.no_nz mt) in
+	let ret = (match m.mt3_ret with
+		| None -> "Void"
+		| Some t -> s_type_path (type_path ctx t)
+	) in
+	let p = ref 0 in
+	let params = List.map (fun a ->
+		let name = (match m.mt3_pnames with
+			| None -> "p" ^ string_of_int !p
+			| Some l -> ident ctx (List.nth l (!p))
+		) in
+		let opt_val = (match m.mt3_dparams with
+			| None -> None
+			| Some l ->
+				try
+					Some (List.nth l (!p - List.length m.mt3_args + List.length l))
+				with
+					_ -> None
+		) in
+		let t = type_val ctx a opt_val in
+		incr p;
+		(if opt_val <> None then "?" else "") ^ name ^ " : " ^ t
+	) m.mt3_args in
+	let vargs = if m.mt3_var_args then " /* ...arguments */" else "" in
+	IO.printf ch "function %s(%s%s) : %s;\n" name (String.concat ", " params) vargs ret
+
+let gen_fields ctx ch fields stat =
+	let fields = List.sort (fun f1 f2 -> compare (ident_rights ctx f1.f3_name) (ident_rights ctx f2.f3_name)) (Array.to_list fields) in
+	List.iter (fun f ->
+		let priv , name = ident_rights ctx f.f3_name in
+		if name.[0] = '$' then
+			()
+		else match f.f3_kind with
+		| A3FMethod m ->
+			if m.m3_override then
+				()
+			else
+			(match m.m3_kind with
+			| MK3Normal ->
+				IO.printf ch "\t";
+				if priv then IO.printf ch "private ";
+				if stat then IO.printf ch "static ";
+				gen_method ctx ch name m.m3_type
+			| MK3Getter ->
+				let set = has_getset fields f m in
+				let set_str = if set then "" else "(default,null)" in
+				let m = As3code.iget ctx.as3_method_types (As3parse.no_nz m.m3_type) in
+				let t = (match m.mt3_ret with None -> "Dynamic" | Some t -> s_type_path (type_path ctx t)) in
+				IO.printf ch "\t%s%svar %s%s : %s;\n" (if priv then "private " else "") (if stat then "static " else "") name set_str t
+			| MK3Setter ->
+				let get = has_getset fields f m in
+				if not get then begin
+					let m = As3code.iget ctx.as3_method_types (As3parse.no_nz m.m3_type) in
+					let t = (match m.mt3_ret with None -> "Dynamic" | Some t -> s_type_path (type_path ctx t)) in
+					IO.printf ch "\t%s%svar %s(null,default) : %s;\n" (if priv then "private " else "") (if stat then "static " else "") name t
+				end;
+			)
+		| A3FVar v ->
+			let t = type_val ctx v.v3_type (Some v.v3_value) in
+			IO.printf ch "\t%s%svar %s : %s;\n" (if priv then "private " else "") (if stat then "static " else "") name t
+		| A3FFunction _ ->
+			assert false
+		| A3FClass _ ->
+			IO.printf ch "\t// ????\n"
+	) fields
+
+let genhx_class ctx c s =
+	let base_path = "hxclasses" in
+	let pack , name = real_type_path ctx c.cl3_name in
+	let skip = (match pack with
+		| [_;x] when String.length x > 3 && String.sub x 0 3 = "as$" -> true
+		| _ when name.[0] = '_' -> true
+		| _ -> false
+	) in
+	if skip then
+		prerr_endline ("// skip " ^ s_type_path (pack,name))
+	else
+	let () = prerr_string ("import " ^ s_type_path (pack,name)) in
+	create_dir "." (base_path :: pack);
+	let f = open_out (base_path ^ "/" ^ (match pack with [] -> "" | l -> String.concat "/" l ^ "/") ^ name ^ ".hx") in
+	let ch = IO.output_channel f in
+	if pack <> [] then IO.printf ch "package %s;\n\n" (String.concat "." pack);
+	IO.printf ch "extern %s %s" (if c.cl3_interface then "interface" else "class") name;
+	let prev = ref (match c.cl3_super with
+	| None -> false
+	| Some p ->
+		match type_path ctx p with
+		| [] , "Dynamic" -> false
+		| path ->
+			IO.printf ch " extends %s" (s_type_path path);
+			true
+	) in
+	Array.iter (fun i ->
+		if !prev then IO.printf ch ",";
+		prev := true;
+		IO.printf ch " implements %s" (s_type_path (type_path ctx i));
+	) c.cl3_implements;
+	IO.printf ch " {\n";
+	IO.printf ch "\t";
+	if not c.cl3_interface then gen_method ctx ch "new" c.cl3_construct;
+	gen_fields ctx ch c.cl3_fields false;
+	gen_fields ctx ch s.st3_fields true;
+	IO.printf ch "}\n";
+	prerr_endline ";";
+	IO.close_out ch
+
+let genhx file =
+	let file = (try Plugin.find_file file with Not_found -> failwith ("File not found : " ^ file)) in
+	let ch = IO.input_channel (open_in_bin file) in
+	SwfParser.full_parsing := true;
+	let _, swf = Swf.parse ch in
+	SwfParser.full_parsing := false;
+	IO.close_in ch;
+	List.iter (fun t ->
+		match t.Swf.tdata with
+		| Swf.TActionScript3 (_,t) -> Array.iteri (fun i c -> genhx_class t c t.as3_statics.(i)) t.as3_classes
+		| _ -> ()
+	) swf
