@@ -25,16 +25,21 @@ type register =
 	| Reg of int
 
 type context = {
+
+	(* segs *)
+	mutable segs : (actions * (string * bool, int) Hashtbl.t) list;
+
 	(* code *)
-	opcodes : actions;
+	mutable opcodes : actions;
 	mutable code_pos : int;
 	mutable stack_size : int;
 	mutable opt_push : bool;
 	mutable ident_count : int;
+	mutable ident_size : int;
 
 	(* management *)
 	packages : (string list,unit) Hashtbl.t;
-	idents : (string * bool,int) Hashtbl.t;
+	mutable idents : (string * bool,int) Hashtbl.t;
 	mutable movieclips : module_path list;
 	mutable inits : texpr list;
 	mutable statics : (tclass * bool * string * texpr) list;
@@ -73,6 +78,7 @@ type kind =
 	| VarStr
 	| VarObj
 	| VarClosure
+	| VarVolatile
 
 type push_style =
 	| VStr of string * bool
@@ -151,6 +157,8 @@ let call ctx kind n =
 			ACall , n + 1
 		| VarClosure | VarObj ->
 			AObjCall , n + 2
+		| VarVolatile ->
+			assert false
 	) in
 	DynArray.add ctx.opcodes op;
 	ctx.opt_push <- false;
@@ -166,6 +174,8 @@ let new_call ctx kind n  =
 			ANew , n + 1
 		| VarClosure | VarObj ->
 			ANewMethod , n + 2
+		| VarVolatile ->
+			assert false
 	) in
 	DynArray.add ctx.opcodes op;
 	ctx.opt_push <- false;
@@ -183,7 +193,7 @@ let rec is_protected_path path ext =
 	| _ -> false
 
 let rec is_protected ctx ?(stat=false) t field =
-	match t with
+	match follow t with
 	| TInst (c,_) ->
 		let rec loop c =
 			(is_protected_path c.cl_path c.cl_extern && PMap.mem field (if stat then c.cl_statics else c.cl_fields))
@@ -191,16 +201,10 @@ let rec is_protected ctx ?(stat=false) t field =
 			|| (not stat && match c.cl_super with None -> false | Some (c,_) -> loop c)
 		in
 		loop c
-	| TMono r ->
-		(match !r with
-		| None -> true (* in Transform.emk only *)
-		| Some t -> is_protected ctx ~stat t field)
-	| TLazy f ->
-		is_protected ctx ~stat ((!f)()) field
-	| TType (t,tl) ->
-		(match t.t_static with
-		| None -> is_protected ctx ~stat (apply_params t.t_types tl t.t_type) field
-		| Some c -> is_protected ctx ~stat:true (TInst (c,[])) field)
+	| TAnon a ->
+		(match !(a.a_status) with
+		| Statics c -> is_protected ctx ~stat:true (TInst (c,[])) field
+		| _ -> false)
 	| _ -> false
 
 let push ctx items =
@@ -213,6 +217,7 @@ let push ctx items =
 			with Not_found ->
 				let n = ctx.ident_count in
 				ctx.ident_count <- n + 1;
+				ctx.ident_size <- ctx.ident_size + 1 + String.length str;
 				Hashtbl.add ctx.idents (str,flag) n;
 				n
 			) in
@@ -286,6 +291,11 @@ let jmp_pos ctx cond =
 		ctx.opt_push <- false
 	)
 
+let init_array ctx n =
+	push ctx [VInt n];
+	write ctx AInitArray;
+	ctx.stack_size <- ctx.stack_size - n
+
 let setvar ?(retval=false) ctx = function
 	| VarReg (-1) -> assert false (** true, false, null **)
 	| VarReg n -> write ctx (ASetReg n); if not retval then write ctx APop
@@ -295,6 +305,11 @@ let setvar ?(retval=false) ctx = function
 		if retval then write ctx (ASetReg 0);
 		write ctx (if s = VarStr then ASet else AObjSet);
 		if retval then push ctx [VReg 0]
+	| VarVolatile ->
+		if retval then write ctx (ASetReg 0);
+		init_array ctx 1;
+		write ctx AObjSet;
+		if retval then push ctx [VReg 0]
 
 let getvar ctx = function
 	| VarReg (-1) -> () (** true, false, null **)
@@ -304,6 +319,10 @@ let getvar ctx = function
 	| VarClosure ->
 		push ctx [VInt 2; VStr ("@closure",false)];
 		call ctx VarStr 2
+	| VarVolatile ->
+		write ctx AObjGet;
+		push ctx [VInt 0];
+		write ctx AObjGet
 
 let gen_path ctx ?(protect=false) (p,t) is_extern =
 	let flag = is_protected_path (p,t) is_extern in
@@ -321,13 +340,13 @@ let gen_path ctx ?(protect=false) (p,t) is_extern =
 		push ctx [VStr (t,flag)];
 		VarObj
 
-let func ctx need_super need_args args =
+let begin_func ctx need_super need_args args =
 	if ctx.version = 6 then
 		let f = {
 			f_name = "";
 			Swf.f_args = List.map snd args;
 			f_codelen = 0;
-		} in
+		} in		
 		write ctx (AFunction f);
 		let start_pos = ctx.code_pos in
 		let old_stack = ctx.fun_stack in
@@ -403,6 +422,15 @@ let free_reg ctx r p =
 	if r <> ctx.reg_count then stack_error p;
 	ctx.reg_count <- ctx.reg_count - 1
 
+let segment ctx =
+	ctx.segs <- (ctx.opcodes,ctx.idents) :: ctx.segs;
+	ctx.opcodes <- DynArray.create();
+	ctx.idents <- Hashtbl.create 0;
+	ctx.ident_count <- 0;
+	ctx.ident_size <- 0;
+	ctx.code_pos <- 0;
+	write ctx (AStringPool [])
+
 (* -------------------------------------------------------------- *)
 (* Generation Helpers *)
 
@@ -547,7 +575,11 @@ let rec gen_access ctx forcall e =
 		push ctx [VStr (f,is_protected ctx e2.etype f)];
 		(match follow e.etype with
 		| TFun _ -> VarClosure
-		| _ -> VarObj)
+		| _ ->
+			if not !protect_all && Transform.is_volatile e.etype then
+				VarVolatile
+			else
+				VarObj)
 	| TArray (ea,eb) ->
 		gen_expr ctx true ea;
 		gen_expr ctx true eb;
@@ -648,7 +680,7 @@ and gen_match ctx retval e cases def =
 	gen_expr ctx true e;
 	let renum = alloc_tmp ctx in
 	set_tmp ctx renum;
-	push ctx [VInt 0];
+	push ctx [VInt 1];
 	write ctx AObjGet;
 	let rtag = alloc_tmp ctx in
 	set_tmp ctx rtag;
@@ -656,7 +688,7 @@ and gen_match ctx retval e cases def =
 	let dispatch = List.map (fun (cl,params,e) ->
 		List.map (fun c ->
 			if !first then first := false else get_tmp ctx rtag;
-			push ctx [VStr (c,false)];
+			push ctx [VInt c];
 			write ctx APhysEqual;
 			cjmp ctx
 		) cl, params, e
@@ -671,7 +703,7 @@ and gen_match ctx retval e cases def =
 		let regs = ctx.regs in
 		let nregs = ctx.reg_count in
 		List.iter (fun j -> j()) jl;
-		let n = ref 0 in
+		let n = ref 1 in
 		List.iter (fun (a,t) ->
 			incr n;
 			match a with
@@ -881,11 +913,8 @@ and gen_expr_2 ctx retval e =
 		) vl;
 		if retval then push ctx [VNull]
 	| TArrayDecl el ->
-		let nitems = List.length el in
 		List.iter (gen_expr ctx true) (List.rev el);
-		push ctx [VInt nitems];
-		write ctx AInitArray;
-		ctx.stack_size <- ctx.stack_size - nitems;
+		init_array ctx (List.length el);
 	| TObjectDecl fl ->
 		let nfields = List.length fl in
 		List.iter (fun (s,v) ->
@@ -926,7 +955,7 @@ and gen_expr_2 ctx retval e =
 				r , ""
 			end
 		) f.tf_args in
-		let tf = func ctx reg_super (Transform.local_find true "__arguments__" f.tf_expr) rargs in
+		let tf = begin_func ctx reg_super (Transform.local_find true "__arguments__" f.tf_expr) rargs in
 		ctx.fun_pargs <- (ctx.code_pos, List.rev !pargs) :: ctx.fun_pargs;
 		if ctx.debug then begin
 			let start_try = gen_try ctx in
@@ -1108,8 +1137,8 @@ let gen_enum_field ctx e f =
 		ctx.reg_count <- 1;
 		let no_reg = ctx.version = 6 in
 		let rargs = List.map (fun (n,_,_) -> if no_reg then 0, n else alloc_reg ctx , "") args in
-		let nregs = List.length rargs + 1 in
-		let tf = func ctx false false rargs in
+		let nregs = List.length rargs + 2 in
+		let tf = begin_func ctx false false rargs in
 		List.iter (fun (r,name) -> 
 			if no_reg then begin
 				push ctx [VStr (name,false)];
@@ -1117,32 +1146,35 @@ let gen_enum_field ctx e f =
 			end else
 				push ctx [VReg r]
 		) (List.rev rargs);
-		push ctx [VStr (f.ef_name,false); VInt nregs];
-		write ctx AInitArray;
+		push ctx [VInt f.ef_index; VStr (f.ef_name,false)];
+		init_array ctx nregs;
+		write ctx ADup;
 		write ctx ADup;
 		push ctx [VStr ("__enum__",false); VThis];
 		write ctx AObjSet;
-		ctx.stack_size <- ctx.stack_size - nregs;
+		push ctx [VStr ("toString",false); VStr ("@estr",false)];
+		write ctx AEval;
+		write ctx AObjSet;
 		write ctx AReturn;
 		tf();
 	| t ->
-		push ctx [VStr (f.ef_name,false); VInt 1];
-		write ctx AInitArray;
-		ctx.stack_size <- ctx.stack_size - 1;
+		push ctx [VInt f.ef_index; VStr (f.ef_name,false)];
+		init_array ctx 2;
+		write ctx ADup;
 		write ctx ADup;
 		push ctx [VStr ("__enum__",false); VReg 0];
 		write ctx AObjSet;
+		push ctx [VStr ("toString",false); VStr ("@estr",false)];
+		write ctx AEval;
+		write ctx AObjSet;
 	);
 	write ctx AObjSet
 
 let init_name ctx path enum =
 	push ctx [VReg 0; VStr ((if enum then "__ename__" else "__name__"),false)];
 	let name = fst path @ [snd path] in
-	let nitems = List.length name in
 	push ctx (List.map (fun s -> VStr (s,false)) (List.rev name));
-	push ctx [VInt nitems];
-	write ctx AInitArray;
-	ctx.stack_size <- ctx.stack_size - nitems;
+	init_array ctx (List.length name);
 	setvar ctx VarObj
 
 let gen_package ctx path ext =
@@ -1186,6 +1218,7 @@ let gen_package ctx path ext =
 	loop [] (fst path)
 
 let gen_type_def ctx t =
+	if ctx.ident_size > 50000 then segment ctx;
 	match t with
 	| TClassDecl c ->
 		(match c.cl_init with
@@ -1215,7 +1248,7 @@ let gen_type_def ctx t =
 			ctx.curmethod <- ("new",false);
 			gen_expr ctx true (Transform.block_vars e)
 		| _ ->
-			let f = func ctx true false [] in
+			let f = begin_func ctx true false [] in
 			f());
 		write ctx (ASetReg 0);
 		setvar ctx acc;
@@ -1254,10 +1287,8 @@ let gen_type_def ctx t =
 			let nimpl = List.length l in
 			push ctx [VReg 0; VStr ("__interfaces__",false)];
 			List.iter (fun (c,_) -> getvar ctx (gen_path ctx c.cl_path c.cl_extern)) l;
-			push ctx [VInt nimpl];
-			write ctx AInitArray;
+			init_array ctx nimpl;
 			setvar ctx VarObj;
-			ctx.stack_size <- ctx.stack_size - nimpl;
 			if ctx.version > 6 then begin
 				List.iter (fun (c,_) -> getvar ctx (gen_path ctx c.cl_path c.cl_extern)) l;
 				push ctx [VInt nimpl; VReg 0];
@@ -1284,6 +1315,10 @@ let gen_type_def ctx t =
 		write ctx (ASetReg 0);
 		setvar ctx acc;
 		init_name ctx e.e_path true;
+		push ctx [VReg 0; VStr ("__constructs__",true)];
+		List.iter (fun s -> push ctx [VStr (s,false)]) (List.rev e.e_names);
+		init_array ctx (List.length e.e_names);
+		write ctx AObjSet;
 		PMap.iter (fun _ f -> gen_enum_field ctx e f) e.e_constrs
 	| TTypeDecl _ ->
 		()
@@ -1331,6 +1366,15 @@ let to_utf8 str =
 			String.iter (fun c -> UTF8.Buf.add_char b (UChar.of_char c)) str;
 			UTF8.Buf.contents b
 
+let build_tag (opcodes,idents) =
+	let idents = Hashtbl.fold (fun ident pos acc -> (ident,pos) :: acc) idents [] in
+	let idents = List.sort (fun (_,p1) (_,p2) -> compare p1 p2) idents in
+	let pidents = List.map (fun ((_,flag),_) -> flag) idents in
+	let idents = AStringPool (List.map (fun ((id,_),_) -> to_utf8 id) idents) in
+	if ActionScript.action_length idents >= 1 lsl 16 then failwith "The SWF can't handle more than a total size of 64K of identifers and literal strings. Try reducing this number by using external data files loaded at runtime";
+	DynArray.set opcodes 0 idents;
+	TDoAction opcodes , pidents
+
 let convert_header ver (w,h,fps,bg) =
 	{
 		h_version = ver;
@@ -1343,7 +1387,7 @@ let convert_header ver (w,h,fps,bg) =
 		};
 		h_frame_count = 1;
 		h_fps = to_float16 fps;
-		h_compressed = true;
+		h_compressed = not (Plugin.defined "no-swf-compress");
 	} , bg
 
 let default_header ver =
@@ -1351,10 +1395,12 @@ let default_header ver =
 
 let generate_code file ver types hres =
 	let ctx = {
+		segs = [];
 		opcodes = DynArray.create();
 		code_pos = 0;
 		stack_size = 0;
 		ident_count = 0;
+		ident_size = 0;
 		opt_push = false;
 		idents = Hashtbl.create 0;
 		packages = Hashtbl.create 0;
@@ -1387,6 +1433,19 @@ let generate_code file ver types hres =
 		write ctx AInitArray;
 		write ctx ASet;
 	end;
+	(* ----- @estr = function() { return flash.Boot.__string_rec(this,""); } ---- *)
+	push ctx [VStr ("@estr",false)];
+	ctx.reg_count <- 1;
+	let f = begin_func ctx false false [] in
+	push ctx [VStr ("xx",false); VThis; VInt 2];
+	getvar ctx (gen_path ctx (["flash"],"Boot") false);
+	push ctx [VStr ("__string_rec",false)]; 
+	call ctx VarObj 2;
+	write ctx AReturn;
+	f();
+	write ctx ASet;
+	ctx.reg_count <- 0;
+	(* ---- *)
 	List.iter (fun t -> gen_type_def ctx t) types;
 	gen_boot ctx hres;
 	List.iter (fun m -> gen_movieclip ctx m) ctx.movieclips;
@@ -1401,14 +1460,11 @@ let generate_code file ver types hres =
 	call ctx VarObj 1;
 	write ctx APop;
 	end_try();
-	let idents = ctx.idents in
-	let idents = Hashtbl.fold (fun ident pos acc -> (ident,pos) :: acc) idents [] in
-	let idents = List.sort (fun (_,p1) (_,p2) -> compare p1 p2) idents in
-	let pidents = List.map (fun ((_,flag),_) -> flag) idents in
-	let idents = AStringPool (List.map (fun ((id,_),_) -> to_utf8 id) idents) in
-	if ActionScript.action_length idents >= 1 lsl 16 then failwith "The SWF can't handle more than a total size of 64K of identifers and literal strings. Try reducing this number by using external data files loaded at runtime";
-	DynArray.set ctx.opcodes 0 idents;
+	let segs = List.rev ((ctx.opcodes,ctx.idents) :: ctx.segs) in
+	let tags = List.map build_tag segs in
 	if Plugin.defined "swf-mark" then begin
+		if List.length segs > 1 then assert false;
+		let pidents = snd (List.hd tags) in
 		let ch = IO.output_channel (open_out_bin (Filename.chop_extension file ^ ".mark")) in
 		IO.write_i32 ch (List.length ctx.fun_pargs);
 		List.iter (fun (id,l) ->
@@ -1420,7 +1476,7 @@ let generate_code file ver types hres =
 		List.iter (fun f -> IO.write_byte ch (if f then 1 else 0)) pidents;
 		IO.close_out ch;
 	end;
-	[TDoAction ctx.opcodes] , ctx.movieclips
+	List.map fst tags , ctx.movieclips
 
 let generate file ver header infile types hres =
 	let t = Plugin.timer "generate swf" in
@@ -1469,7 +1525,13 @@ let generate file ver header infile types hres =
 			[]
 	in
 	let sandbox() =
-		tag (TSandbox (if ver = 9 then SBUnknown 8 else SBLocal))
+		let net = Plugin.defined "network-sandbox" in
+		tag (TSandbox (match ver, net with
+			| 9, true -> SBUnknown 9
+			| 9, false -> SBUnknown 8
+			| _, true -> SBNetwork
+			| _, false -> SBLocal
+		))
 	in
 	let swf = (match infile with
 		| None ->
@@ -1490,7 +1552,7 @@ let generate file ver header infile types hres =
 					h , Some tagbg
 			) in
 			IO.close_in ch;
-			let no_sandbox = ref true in
+			let has_sandbox = ref false in
 			let rec loop = function
 				| [] ->
 					failwith ("Frame 1 not found in " ^ file)
@@ -1519,14 +1581,17 @@ let generate file ver header infile types hres =
 						) el;
 						t :: loop l
 					end;
-				| ({ tdata = TSandbox s } as t) :: l when ver = 9 ->
-					no_sandbox := false;
+				| ({ tdata = TSandbox _ } as t) :: l when ver = 9 ->
+					has_sandbox := true;
 					{ t with tdata = TSandbox (SBUnknown 8) } :: loop l
+				| ({ tdata = TSandbox _ } as t) :: l ->				
+					has_sandbox := true;
+					t :: loop l
 				| t :: l ->
 					t :: loop l
 			in
 			let tags = loop swf in
-			let tags = (if !no_sandbox && ver >= 8 then sandbox() :: tags else tags) in
+			let tags = (if not !has_sandbox && ver >= 8 then sandbox() :: tags else tags) in
 			(header , tags)
 	) in
 	let swf = if ver = 8 && Plugin.defined "flash_v9" then ({ (fst swf) with h_version = 9 }, snd swf) else swf in
