@@ -51,6 +51,7 @@ type register = {
 type 'a access =
 	| VReg of register
 	| VId of as3_name
+	| VCast of as3_name * tkind
 	| VGlobal of as3_name
 	| VArray
 	| VScope of as3_slot
@@ -129,9 +130,11 @@ let t_void = TEnum ({
 		e_extern = false;
 		e_types = [];
 		e_constrs = PMap.empty;
+		e_names = [];
 	},[])
 
 let t_string = TInst (mk_class ([],"String") null_pos None false,[])
+let t_int = TInst (mk_class ([],"Int") null_pos None false,[])
 
 let lookup i w =
 	try
@@ -193,6 +196,7 @@ let type_path ctx ?(getclass=false) path =
 		| [] , "Float" -> [] , "Number"
 		| [] , "Bool" -> [] , "Boolean"
 		| [] , "Void" -> [] , "void"
+		| [] , "Enum" -> [] , "Class"
 		| ["flash"] , "FlashXml__" -> [] , "Xml"
 		| ["flash"] , "Boot" -> [] , ctx.boot
 		| _ -> path
@@ -263,6 +267,26 @@ let classify ctx t =
 
 let ident ctx i = type_path ctx ([],i)
 
+let as3 ctx p =
+	let pid = string ctx p in
+	let ns = string ctx "http://adobe.com/AS3/2006/builtin" in
+	let nsid = lookup (A3NNamespace ns) ctx.namespaces in
+	let tid = lookup (A3MName (pid,nsid)) ctx.names in
+	tid
+
+let property ctx p t =
+	match follow t with
+	| TInst ({ cl_path = [],"Array" },_) ->
+		(match p with
+		| "length" | "copy" | "insert" | "remove" | "iterator" -> ident ctx p
+		| _ -> as3 ctx p);
+	| TInst ({ cl_path = [],"String" },_) ->
+		(match p with
+		| "length" | "charCodeAt" (* use haXe version *) -> ident ctx p
+		| _ -> as3 ctx p);
+	| _ ->
+		ident ctx p
+
 let default_infos() =
 	{
 		ipos = 0;
@@ -371,18 +395,18 @@ let rec setvar ctx (acc : write access) retval =
 	| VGlobal g ->
 		if retval then write ctx A3Dup;
 		write ctx (A3SetProp g);
-	| VId _ | VArray | VScope _ when retval ->
+	| VId _ | VCast _ | VArray | VScope _ when retval ->
 		let r = alloc_reg ctx KDynamic in
 		write ctx A3Dup;
 		set_reg ctx r;
 		setvar ctx acc false;
 		write ctx (A3Reg r.rid);
 		free_reg ctx r
-	| VId id ->
+	| VId id | VCast (id,_) ->
 		write ctx (A3InitProp id)
 	| VArray ->
 		let id_aset = lookup (A3MMultiNameLate ctx.gpublic) ctx.names in
-		write ctx (A3InitProp id_aset);
+		write ctx (A3SetProp id_aset);
 		ctx.infos.istack <- ctx.infos.istack - 1
 	| VScope n ->
 		write ctx (A3SetSlot n)
@@ -390,9 +414,16 @@ let rec setvar ctx (acc : write access) retval =
 let getvar ctx (acc : read access) =
 	match acc with
 	| VReg r ->
+		if not r.rinit then begin
+			r.rinit <- true;
+			r.rcond <- true;
+		end;
 		write ctx (A3Reg r.rid)
 	| VId id ->
 		write ctx (A3GetProp id)
+	| VCast (id,t) ->
+		write ctx (A3GetProp id);
+		coerce ctx t
 	| VGlobal g ->
 		write ctx (A3GetLex g);
 	| VArray ->
@@ -630,12 +661,15 @@ let gen_access ctx e (forset : 'a) : 'a access =
 	match e.eexpr with
 	| TLocal i ->
 		gen_local_access ctx i e.epos forset
-	| TField (e,f) ->
-		let id = ident ctx f in
-		(match e.eexpr with
+	| TField (e1,f) ->
+		let id = property ctx f e1.etype in
+		(match e1.eexpr with
 		| TConst TThis when not ctx.in_static -> write ctx (A3FindPropStrict id)
-		| _ -> gen_expr ctx true e);
-		VId id
+		| _ -> gen_expr ctx true e1);
+		(match follow e1.etype with
+		| TInst _ | TEnum _ -> VId id
+		| TAnon a when (match !(a.a_status) with Statics _ | EnumStatics _ -> true | _ -> false) -> VId id
+		| _ -> VCast (id,classify ctx e.etype))
 	| TArray ({ eexpr = TLocal "__global__" },{ eexpr = TConst (TString s) }) ->
 		let path = (match List.rev (ExtString.String.nsplit s ".") with [] -> assert false | x :: l -> List.rev l, x) in
 		let id = type_path ctx path in
@@ -715,7 +749,7 @@ let rec gen_expr_content ctx retval e =
 		getvar ctx (gen_access ctx e Read);
 		coerce ctx (classify ctx e.etype)
 	| TBinop (op,e1,e2) ->
-		gen_binop ctx retval op e1 e2
+		gen_binop ctx retval op e1 e2 e.etype
 	| TCall (e,el) ->
 		gen_call ctx e el
 	| TNew (c,_,pl) ->
@@ -888,11 +922,11 @@ let rec gen_expr_content ctx retval e =
 	| TMatch (e0,_,cases,def) ->
 		let t = classify ctx e.etype in
 		let rparams = alloc_reg ctx KDynamic in
-		let rtag = alloc_reg ctx KDynamic in
+		let rindex = alloc_reg ctx KInt in
 		gen_expr ctx true e0;
 		write ctx A3Dup;
-		write ctx (A3GetProp (ident ctx "tag"));
-		set_reg ctx rtag;
+		write ctx (A3GetProp (ident ctx "index"));
+		set_reg ctx rindex;
 		write ctx (A3GetProp (ident ctx "params"));
 		set_reg ctx rparams;
 		let branch = begin_branch ctx in
@@ -903,12 +937,12 @@ let rec gen_expr_content ctx retval e =
 				| [] ->
 					assert false
 				| [tag] ->
-					write ctx (A3Reg rtag.rid);
-					write ctx (A3String (lookup tag ctx.strings));
+					write ctx (A3Reg rindex.rid);
+					write ctx (A3Int tag);
 					prev := jump ctx J3Neq;
 				| tag :: l ->
-					write ctx (A3Reg rtag.rid);
-					write ctx (A3String (lookup tag ctx.strings));
+					write ctx (A3Reg rindex.rid);
+					write ctx (A3Int tag);
 					let j = jump ctx J3Eq in
 					loop l;
 					j()
@@ -942,18 +976,18 @@ let rec gen_expr_content ctx retval e =
 		) cases in
 		(!prev)();
 		(match def with
-		| None -> 
+		| None ->
 			if retval then begin
 				write ctx A3Null;
 				coerce ctx t;
 			end;
-		| Some e -> 
+		| Some e ->
 			gen_expr ctx retval e;
 			if retval && classify ctx e.etype <> t then coerce ctx t;
 		);
 		List.iter (fun j -> j()) jend;
 		branch();
-		free_reg ctx rtag;
+		free_reg ctx rindex;
 		free_reg ctx rparams
 
 and gen_call ctx e el =
@@ -996,7 +1030,7 @@ and gen_call ctx e el =
 	| TLocal "__delete__" , [o;f] ->
 		gen_expr ctx true o;
 		gen_expr ctx true f;
-		write ctx (A3DeleteProp (lookup (A3MMultiNameLate ctx.gpublic) ctx.names));		
+		write ctx (A3DeleteProp (lookup (A3MMultiNameLate ctx.gpublic) ctx.names));
 	| TLocal "__unprotect__" , [e] ->
 		gen_expr ctx true e
 	| TLocal "__typeof__", [e] ->
@@ -1027,10 +1061,31 @@ and gen_call ctx e el =
 		write ctx (A3FindPropStrict id);
 		List.iter (gen_expr ctx true) el;
 		write ctx (A3CallProperty (id,List.length el));
-	| TField (e,f) , _ ->
-		gen_expr ctx true e;
+	| TField (e1,f) , _ ->
+		gen_expr ctx true e1;
 		List.iter (gen_expr ctx true) el;
-		write ctx (A3CallProperty (ident ctx f,List.length el));
+		write ctx (A3CallProperty (property ctx f e1.etype,List.length el));
+		let coerce() =
+			match follow e.etype with
+			| TFun (_,r) -> coerce ctx (classify ctx r)
+			| _ -> ()
+		in
+		(match follow e1.etype with
+		| TInst ({ cl_path = [],"Array" },_) -> 
+			(match f with
+			| "copy" | "remove" -> coerce()
+			| _ -> ())
+		| TInst ({ cl_path = [],"Date" },_) -> 
+			coerce() (* all date methods are typed as Number in AS3 and Int in haXe *) 
+		| TAnon a when (match !(a.a_status) with Statics { cl_path = ([],"Date") } -> true | _ -> false) ->
+			(match f with
+			| "now" | "fromString" | "fromTime"  -> coerce()
+			| _ -> ())
+		| TAnon a when (match !(a.a_status) with Statics { cl_path = ([],"Math") } -> true | _ -> false) ->
+			(match f with
+			| "isFinite" | "isNaN" -> coerce()
+			| _ -> ())
+		| _ -> ())
 	| TEnumField (e,f) , _ ->
 		let id = type_path ctx ~getclass:true e.e_path in
 		write ctx (A3GetLex id);
@@ -1060,6 +1115,7 @@ and gen_unop ctx retval op flag e =
 		let acc = gen_access ctx e Write in (* for set *)
 		match acc with
 		| VReg r when not retval && r.rtype = KInt ->
+			if not r.rinit then r.rcond <- true;
 			write ctx (if incr then A3IncrIReg r.rid else A3DecrIReg r.rid)
 		| _ ->
 		getvar ctx (gen_access ctx e Read);
@@ -1076,13 +1132,20 @@ and gen_unop ctx retval op flag e =
 			write ctx (A3Op (if incr then A3OIncr else A3ODecr));
 			setvar ctx acc retval
 
-and gen_binop ctx retval op e1 e2 =
+and gen_binop ctx retval op e1 e2 t =
 	let gen_op ?iop o =
 		gen_expr ctx true e1;
 		gen_expr ctx true e2;
 		match iop with
-		| Some iop when classify ctx e1.etype = KInt && classify ctx e2.etype = KInt ->
-			write ctx (A3Op iop)
+		| Some iop -> 
+			let k1 = classify ctx e1.etype in
+			let k2 = classify ctx e2.etype in
+			if k1 = KInt && k2 = KInt then
+				write ctx (A3Op iop)
+			else begin
+				write ctx (A3Op o);
+				if o = A3OAdd then coerce ctx (classify ctx t);
+			end;
 		| _ ->
 			write ctx (A3Op o)
 	in
@@ -1109,7 +1172,7 @@ and gen_binop ctx retval op e1 e2 =
 		j();
 	| OpAssignOp op ->
 		let acc = gen_access ctx e1 Write in
-		gen_binop ctx true op e1 e2;
+		gen_binop ctx true op e1 e2 t;
 		setvar ctx acc retval
 	| OpAdd ->
 		gen_op ~iop:A3OIAdd A3OAdd
@@ -1221,12 +1284,12 @@ let generate_class_init ctx c slot =
 let generate_class_statics ctx c =
 	let r = alloc_reg ctx KDynamic in
 	let first = ref true in
-	let nslot = ref 1 in
+	let nslot = ref 0 in
 	List.iter (fun f ->
-		incr nslot;
 		match f.cf_expr with
-		| Some { eexpr = TFunction _ } | None -> ()
+		| Some { eexpr = TFunction _ } when f.cf_set <> NormalAccess -> ()
 		| Some e ->
+			incr nslot;
 			if !first then begin
 				write ctx A3GetGlobalScope;
 				write ctx (A3GetProp (type_path ctx c.cl_path));
@@ -1236,6 +1299,8 @@ let generate_class_statics ctx c =
 			write ctx (A3Reg r.rid);
 			gen_expr ctx true (Transform.block_vars e);
 			write ctx (A3SetSlot !nslot);
+		| _ ->
+			incr nslot
 	) c.cl_ordered_statics;
 	free_reg ctx r
 
@@ -1261,10 +1326,15 @@ let generate_enum_init ctx e slot =
 			write ctx (A3Reg r.rid);
 			write ctx (A3FindPropStrict name_id);
 			write ctx (A3String (lookup f.ef_name ctx.strings));
+			write ctx (A3Int f.ef_index);
 			write ctx A3Null;
-			write ctx (A3ConstructProperty (name_id,2));
+			write ctx (A3ConstructProperty (name_id,3));
 			write ctx (A3SetSlot !nslot);
 	) e.e_constrs;
+	write ctx (A3Reg r.rid);
+	List.iter (fun n -> write ctx (A3String (lookup n ctx.strings))) e.e_names;
+	write ctx (A3Array (List.length e.e_names));
+	write ctx (A3SetSlot (!nslot + 1));
 	free_reg ctx r
 
 let generate_field_kind ctx f c stat =
@@ -1349,7 +1419,7 @@ let generate_class ctx c =
 	let sc = {
 		cl3_name = name_id;
 		cl3_super = (if c.cl_interface then None else Some (type_path ctx (match c.cl_super with None -> [],"Object" | Some (c,_) -> c.cl_path)));
-		cl3_sealed = c.cl_path <> (["flash"],"Boot");
+		cl3_sealed = c.cl_path <> (["flash"],"Boot") && c.cl_dynamic = None;
 		cl3_final = false;
 		cl3_interface = c.cl_interface;
 		cl3_namespace = None;
@@ -1360,15 +1430,18 @@ let generate_class ctx c =
 		cl3_construct = cid;
 		cl3_fields = fields;
 	} in
-	let st_count = ref 1 in
+	let st_field_count = ref 0 in
+	let st_meth_count = ref 0 in
 	let st = {
 		st3_method = st_id;
 		st3_fields = Array.of_list (List.map (fun f ->
-			incr st_count;
+			let k = (match generate_field_kind ctx f c true with None -> assert false | Some k -> k) in
+			let count = (match k with A3FMethod _ -> st_meth_count | A3FVar _ -> st_field_count | _ -> assert false) in
+			incr count;
 			{
 				f3_name = ident ctx f.cf_name;
-				f3_slot = !st_count;
-				f3_kind = (match generate_field_kind ctx f c true with None -> assert false | Some k -> k);
+				f3_slot = !count;
+				f3_kind = k;
 				f3_metas = None;
 			}
 		) c.cl_ordered_statics)
@@ -1379,14 +1452,18 @@ let generate_class ctx c =
 let generate_enum ctx e =
 	let name_id = type_path ctx e.e_path in
 	let st_id = empty_method ctx in
-	let f = begin_fun ctx [("tag",false,t_string);("params",false,mk_mono())] t_void [] false in
+	let f = begin_fun ctx [("tag",false,t_string);("index",false,t_int);("params",false,mk_mono())] t_void [] false in
 	let tag_id = ident ctx "tag" in
+	let index_id = ident ctx "index" in
 	let params_id = ident ctx "params" in
 	write ctx (A3FindProp tag_id);
 	write ctx (A3Reg 1);
 	write ctx (A3InitProp tag_id);
-	write ctx (A3FindProp params_id);
+	write ctx (A3FindProp index_id);
 	write ctx (A3Reg 2);
+	write ctx (A3InitProp index_id);
+	write ctx (A3FindProp params_id);
+	write ctx (A3Reg 3);
 	write ctx (A3InitProp params_id);
 	write ctx A3RetVoid;
 	let construct = f() in
@@ -1407,6 +1484,7 @@ let generate_enum ctx e =
 		cl3_construct = construct;
 		cl3_fields = [|
 			{ f3_name = tag_id; f3_slot = 0; f3_kind = A3FVar { v3_type = None; v3_value = A3VNone; v3_const = false; }; f3_metas = None };
+			{ f3_name = index_id; f3_slot = 0; f3_kind = A3FVar { v3_type = None; v3_value = A3VNone; v3_const = false; }; f3_metas = None };
 			{ f3_name = params_id; f3_slot = 0; f3_kind = A3FVar { v3_type = None; v3_value = A3VNone; v3_const = false; }; f3_metas = None };
 			{ f3_name = ident ctx "__enum__"; f3_slot = 0; f3_kind = A3FVar { v3_type = None; v3_value = A3VBool true; v3_const = true }; f3_metas = None };
 			{
@@ -1433,10 +1511,11 @@ let generate_enum ctx e =
 					let fdata = begin_fun ctx args (TEnum (e,[])) [] true in
 					write ctx (A3FindPropStrict name_id);
 					write ctx (A3String (lookup f.ef_name ctx.strings));
+					write ctx (A3Int f.ef_index);
 					let n = ref 0 in
 					List.iter (fun _ -> incr n; write ctx (A3Reg !n)) args;
 					write ctx (A3Array (!n));
-					write ctx (A3ConstructProperty (name_id,2));
+					write ctx (A3ConstructProperty (name_id,3));
 					write ctx A3Ret;
 					let fid = fdata() in
 					A3FMethod {
@@ -1455,9 +1534,14 @@ let generate_enum ctx e =
 		st3_method = st_id;
 		st3_fields = Array.of_list ({
 			f3_name = ident ctx "__isenum";
-			f3_slot = !st_count + 1;
+			f3_slot = !st_count + 2;
 			f3_kind = A3FVar { v3_type = None; v3_value = A3VBool true; v3_const = true; };
 			f3_metas = None;
+		} :: {
+			f3_name = ident ctx "__constructs__";
+			f3_slot = !st_count + 1;
+			f3_kind = A3FVar { v3_type = None; v3_value = A3VNone; v3_const = false; };
+			f3_metas = None;
 		} :: constrs)
 	} in
 	ctx.classes <- sc :: ctx.classes;
@@ -1473,7 +1557,7 @@ let generate_resources ctx hres =
 	write ctx A3GetGlobalScope;
 	write ctx (A3GetProp (type_path ctx ([],ctx.boot)));
 	let id = type_path ctx (["flash";"utils"],"Dictionary") in
-	write ctx (A3FindPropStrict id);	
+	write ctx (A3FindPropStrict id);
 	write ctx (A3ConstructProperty (id,0));
 	let r = alloc_reg ctx (KType id) in
 	set_reg ctx r;
@@ -1549,6 +1633,7 @@ let generate types hres =
 	let nsets = new_lookup() in
 	let empty_id = lookup "" strings in
 	let rpublic = lookup (A3NPublic (Some empty_id)) namespaces in
+	Random.self_init();
 	let ctx = {
 		boot = "Boot_" ^ Printf.sprintf "%X" (Random.int 0xFFFFFF);
 		strings = strings;
@@ -1576,7 +1661,7 @@ let generate types hres =
 		last_line = -1;
 		try_scope_reg = None;
 	} in
-	List.iter (generate_type ctx) types;	
+	List.iter (generate_type ctx) types;
 	let init = generate_inits ctx types hres in
 	let a = {
 		as3_ints = lookup_array ctx.ints;
@@ -1596,232 +1681,5 @@ let generate types hres =
 	} in
 	[Swf.TActionScript3 (Some (0,""),a)], ctx.boot
 
-
-(* ----------------------------------------------------------------------------------------
-
-	HX generation
-
-   ---------------------------------------------------------------------------------------- *)
-
-let ident ctx p =
-	As3code.iget ctx.as3_idents p
-
-let package ctx idx =
-	match As3code.iget ctx.as3_namespaces idx with
-	| A3NPrivate (Some id)
-	| A3NPublic (Some id)
-	| A3NInternal (Some id)
-	| A3NProtected id
-	| A3NExplicit id
-	| A3NStaticProtected (Some id) ->
-		let pack = ident ctx id in
-		ExtString.String.nsplit pack "."
-	| A3NNamespace id ->
-		["/* namespace " ^ ident ctx id ^ "*/"]
-	| A3NPrivate None | A3NPublic None | A3NInternal None | A3NStaticProtected None ->
-		[]
-
-let rec real_type_path ctx p =
-	let rec loop = function
-		| A3MName (id,pack) ->
-			let name = ident ctx id in
-			let pack = package ctx pack in
-			pack , name
-		| A3MMultiName (Some id,pack) ->
-			let name = ident ctx id in
-			let pack = package ctx (List.hd (As3code.iget ctx.as3_nsets pack)) in
-			pack , name
-		| A3MMultiName (None,_) ->
-			[] , "$MultiName"
-		| A3MMultiNameLate _ ->
-			[] , "$MultiNameLate"
-		| A3MRuntimeName _ ->
-			[] , "$RuntimeName"
-		| A3MRuntimeNameLate ->
-			[] , "$RuntimeNameLate"
-		| A3MAttrib n ->
-			let path, name = loop n in
-			"$Attrib" :: path, name
-	in
-	loop (As3code.iget ctx.as3_names p)
-
-let type_path ctx p =
-	match real_type_path ctx p with
-	| [] , "Object" -> [] , "Dynamic"
-	| [] , "Boolean" -> [] , "Bool"
-	| [] , "int" -> [] , "Int"
-	| [] , "uint" -> [] , "UInt"
-	| [] , "Number" -> [] , "Float"
-	| [] , "Array" -> [] , "Array<Dynamic>"
-	| [] , "void" -> [] , "Void"
-	| [] , "Function" -> [] , "Dynamic"
-	| [] , "Class" -> [] , "Class<Dynamic>"
-	| path -> path
-
-let ident_rights ctx id =
-	match As3code.iget ctx.as3_names id with
-	| A3MName (id,r) ->
-		let name = ident ctx id in
-		(match As3code.iget ctx.as3_namespaces r with
-		| A3NNamespace i when As3code.iget ctx.as3_idents i = "http://www.adobe.com/2006/flex/mx/internal" -> false, "$" ^ name
-		| A3NPublic _ | A3NNamespace _ -> false , name
-		| _ -> true , name)
-	| _ -> false, "???"
-
-let rec create_dir acc = function
-	| [] -> ()
-	| d :: l ->
-		let path = acc ^ "/" ^ d in
-		(try Unix.mkdir path 0o777 with _ -> ());
-		create_dir path l
-
-let value_type = function
-	| A3VNone
-	| A3VNull -> "Dynamic"
-	| A3VBool _ -> "Bool"
-	| A3VString _ -> "String"
-	| A3VInt _ -> "Int"
-	| A3VUInt _ -> "UInt"
-	| A3VFloat _ -> "Float"
-	| A3VNamespace _ -> "$Namespace"
-
-let type_val ctx t v =
-	match t with
-	| None ->
-		(match v with
-		| None -> "Dynamic"
-		| Some v -> value_type v)
-	| Some t ->
-		s_type_path (type_path ctx t)
-
-let has_getset ml f m =
-	List.exists (fun f2 ->
-		match f2.f3_kind with
-		| A3FMethod m2 when f.f3_name = f2.f3_name ->
-			(match m.m3_kind , m2.m3_kind with
-			| MK3Getter , MK3Setter | MK3Setter , MK3Getter -> true
-			| _ -> false)
-		| _ -> false
-	) ml
-
-let gen_method ctx ch name mt =
-	let m = As3code.iget ctx.as3_method_types (As3parse.no_nz mt) in
-	let ret = (match m.mt3_ret with
-		| None -> "Void"
-		| Some t -> s_type_path (type_path ctx t)
-	) in
-	let p = ref 0 in
-	let params = List.map (fun a ->
-		let name = (match m.mt3_pnames with
-			| None -> "p" ^ string_of_int !p
-			| Some l -> ident ctx (List.nth l (!p))
-		) in
-		let opt_val = (match m.mt3_dparams with
-			| None -> None
-			| Some l ->
-				try
-					Some (List.nth l (!p - List.length m.mt3_args + List.length l))
-				with
-					_ -> None
-		) in
-		let t = type_val ctx a opt_val in
-		incr p;
-		(if opt_val <> None then "?" else "") ^ name ^ " : " ^ t
-	) m.mt3_args in
-	let vargs = if m.mt3_var_args then " /* ...arguments */" else "" in
-	IO.printf ch "function %s(%s%s) : %s;\n" name (String.concat ", " params) vargs ret
-
-let gen_fields ctx ch fields stat =
-	let fields = List.sort (fun f1 f2 -> compare (ident_rights ctx f1.f3_name) (ident_rights ctx f2.f3_name)) (Array.to_list fields) in
-	List.iter (fun f ->
-		let priv , name = ident_rights ctx f.f3_name in
-		if name.[0] = '$' then
-			()
-		else match f.f3_kind with
-		| A3FMethod m ->
-			if m.m3_override then
-				()
-			else
-			(match m.m3_kind with
-			| MK3Normal ->
-				IO.printf ch "\t";
-				if priv then IO.printf ch "private ";
-				if stat then IO.printf ch "static ";
-				gen_method ctx ch name m.m3_type
-			| MK3Getter ->
-				let set = has_getset fields f m in
-				let set_str = if set then "" else "(default,null)" in
-				let m = As3code.iget ctx.as3_method_types (As3parse.no_nz m.m3_type) in
-				let t = (match m.mt3_ret with None -> "Dynamic" | Some t -> s_type_path (type_path ctx t)) in
-				IO.printf ch "\t%s%svar %s%s : %s;\n" (if priv then "private " else "") (if stat then "static " else "") name set_str t
-			| MK3Setter ->
-				let get = has_getset fields f m in
-				if not get then begin
-					let m = As3code.iget ctx.as3_method_types (As3parse.no_nz m.m3_type) in
-					let t = (match m.mt3_ret with None -> "Dynamic" | Some t -> s_type_path (type_path ctx t)) in
-					IO.printf ch "\t%s%svar %s(null,default) : %s;\n" (if priv then "private " else "") (if stat then "static " else "") name t
-				end;
-			)
-		| A3FVar v ->
-			let t = type_val ctx v.v3_type (Some v.v3_value) in
-			IO.printf ch "\t%s%svar %s : %s;\n" (if priv then "private " else "") (if stat then "static " else "") name t
-		| A3FFunction _ ->
-			assert false
-		| A3FClass _ ->
-			IO.printf ch "\t// ????\n"
-	) fields
-
-let genhx_class ctx c s =
-	let base_path = "hxclasses" in
-	let pack , name = real_type_path ctx c.cl3_name in
-	let skip = (match pack with
-		| [_;x] when String.length x > 3 && String.sub x 0 3 = "as$" -> true
-		| _ when name.[0] = '_' -> true
-		| _ -> false
-	) in
-	if skip then
-		prerr_endline ("// skip " ^ s_type_path (pack,name))
-	else
-	let () = prerr_string ("import " ^ s_type_path (pack,name)) in
-	create_dir "." (base_path :: pack);
-	let f = open_out (base_path ^ "/" ^ (match pack with [] -> "" | l -> String.concat "/" l ^ "/") ^ name ^ ".hx") in
-	let ch = IO.output_channel f in
-	if pack <> [] then IO.printf ch "package %s;\n\n" (String.concat "." pack);
-	IO.printf ch "extern %s %s" (if c.cl3_interface then "interface" else "class") name;
-	let prev = ref (match c.cl3_super with
-	| None -> false
-	| Some p ->
-		match type_path ctx p with
-		| [] , "Dynamic" -> false
-		| path ->
-			IO.printf ch " extends %s" (s_type_path path);
-			true
-	) in
-	Array.iter (fun i ->
-		if !prev then IO.printf ch ",";
-		prev := true;
-		IO.printf ch " implements %s" (s_type_path (type_path ctx i));
-	) c.cl3_implements;
-	IO.printf ch " {\n";
-	IO.printf ch "\t"; gen_method ctx ch "new" c.cl3_construct;
-	gen_fields ctx ch c.cl3_fields false;
-	gen_fields ctx ch s.st3_fields true;
-	IO.printf ch "}\n";
-	prerr_endline ";";
-	IO.close_out ch
-
-let genhx file =
-	let file = (try Plugin.find_file file with Not_found -> failwith ("File not found : " ^ file)) in
-	let ch = IO.input_channel (open_in_bin file) in
-	SwfParser.full_parsing := true;
-	let _, swf = Swf.parse ch in
-	SwfParser.full_parsing := false;
-	IO.close_in ch;
-	List.iter (fun t ->
-		match t.Swf.tdata with
-		| Swf.TActionScript3 (_,t) -> Array.iteri (fun i c -> genhx_class t c t.as3_statics.(i)) t.as3_classes
-		| _ -> ()
-	) swf
-
 ;;
 gen_expr_ref := gen_expr
