@@ -38,6 +38,7 @@ let get_full_path file =
 	Extc.get_full_path file
 
 let normalize_path p =
+	let p = ExtString.String.strip p in
 	let l = String.length p in
 	if l = 0 then
 		"./"
@@ -129,7 +130,7 @@ let read_type_path p cp =
 	let p = (match p with "flash" :: l when Plugin.defined "flash9" -> "flash9" :: l | _ -> p) in
 	List.iter (fun path ->
 		let dir = path ^ String.concat "/" p in
-		let r = Sys.readdir dir in
+		let r = (try Sys.readdir dir with _ -> [||]) in
 		Array.iter (fun f ->
 			if (Unix.stat (dir ^ "/" ^ f)).Unix.st_kind = Unix.S_DIR then begin
 				if f.[0] > 'a' && f.[0] < 'z' then packages := f :: !packages
@@ -158,7 +159,7 @@ let rec process_params acc = function
 
 and init params =
 try
-	let version = 114 in
+	let version = 116 in
 	let version_str = Printf.sprintf "%d.%.2d" (version / 100) (version mod 100) in
 	let usage = "Haxe Compiler " ^ version_str ^ " - (c)2005-2007 Motion-Twin\n Usage : haxe.exe [options] <class names...>\n Options :" in
 	let classes = ref [([],"Std")] in
@@ -273,12 +274,16 @@ try
 			Plugin.define def;
 		),"<var> : define a conditional compilation flag");
 		("-resource",Arg.String (fun res ->
-			match ExtString.String.nsplit res "@" with
+			match ExtString.String.nsplit (ExtString.String.strip res) "@" with
 			| [file; name] ->
 				let file = (try Plugin.find_file file with Not_found -> file) in
 				let data = Std.input_file ~bin:true file in
 				if Hashtbl.mem hres name then failwith ("Duplicate resource name " ^ name);
 				Hashtbl.add hres name data
+			| [file] ->
+				let file = (try Plugin.find_file file with Not_found -> file) in
+				let data = Std.input_file ~bin:true file in
+				Hashtbl.replace hres file data
 			| _ ->
 				raise (Arg.Bad "Invalid Resource format : should be file@name")
 		),"<file@name> : add a named resource file");
@@ -310,7 +315,7 @@ try
 		("--neko-source", define "neko_source", ": keep generated neko source");
 		("--gen-hx-classes", Arg.String (fun file ->
 			gen_hx := true;
-			Genswf9.genhx file
+			Genas3.genhx file
 		),"<file> : generate hx headers from SWF9 file");
 		("--next", Arg.Unit (fun() -> assert false), ": separate several haxe compilations");
 		("--altfmt", Arg.Unit (fun() -> alt_format := true),": use alternative error output format");
