@@ -27,11 +27,6 @@ type field_access =
 	| MethodAccess of string
 	| F9MethodAccess
 
-type anon_status =
-	| Closed
-	| Opened
-	| Statics
-
 type t =
 	| TMono of t option ref
 	| TEnum of tenum * tparams
@@ -59,6 +54,13 @@ and tfunc = {
 	tf_expr : texpr;
 }
 
+and anon_status =
+	| Closed
+	| Opened
+	| Const
+	| Statics of tclass
+	| EnumStatics of tenum
+
 and tanon = {
 	mutable a_fields : (string, tclass_field) PMap.t;
 	a_status : anon_status ref;
@@ -85,7 +87,7 @@ and texpr_expr =
 	| TIf of texpr * texpr * texpr option
 	| TWhile of texpr * texpr * Ast.while_flag
 	| TSwitch of texpr * (texpr list * texpr) list * texpr option
-	| TMatch of texpr * (tenum * tparams) * (string list * (string option * t) list option * texpr) list * texpr option
+	| TMatch of texpr * (tenum * tparams) * (int list * (string option * t) list option * texpr) list * texpr option
 	| TTry of texpr * (string * t * texpr) list
 	| TReturn of texpr option
 	| TBreak
@@ -135,6 +137,7 @@ and tenum_field = {
 	ef_type : t;
 	ef_pos : Ast.pos;
 	ef_doc : Ast.documentation;
+	ef_index : int;
 }
 
 and tenum = {
@@ -145,6 +148,7 @@ and tenum = {
 	e_extern : bool;
 	mutable e_types : (string * t) list;
 	mutable e_constrs : (string , tenum_field) PMap.t;
+	mutable e_names : string list;
 }
 
 and tdef = {
@@ -152,7 +156,6 @@ and tdef = {
 	t_pos : Ast.pos;
 	t_doc : Ast.documentation;
 	t_private : bool;
-	t_static : tclass option;
 	mutable t_types : (string * t) list;
 	mutable t_type : t;
 }
@@ -171,7 +174,6 @@ type module_def = {
 let mk e t p = { eexpr = e; etype = t; epos = p }
 
 let not_opened = ref Closed
-let static_status = ref Statics
 let is_closed a = !(a.a_status) <> Opened
 let mk_anon fl = TAnon { a_fields = fl; a_status = not_opened; }
 
@@ -416,6 +418,7 @@ type unify_error =
 	| Cannot_unify of t * t
 	| Invalid_field_type of string
 	| Has_no_field of t * string
+	| Has_extra_field of t * string
 	| Invalid_access of string * bool * field_access * field_access
 	| Invalid_visibility of string
 	| Not_matching_optional of string
@@ -428,6 +431,7 @@ let invalid_field n = Invalid_field_type n
 let invalid_access n get a b = Invalid_access (n,get,a,b)
 let invalid_visibility n = Invalid_visibility n
 let has_no_field t n = Has_no_field (t,n)
+let has_extra_field t n = Has_extra_field (t,n)
 let error l = raise (Unify_error l)
 
 let unify_access a1 a2 =
@@ -574,8 +578,6 @@ let rec unify a b =
 		(match !t with
 		| None -> if not (link t b a) then error [cannot_unify a b]
 		| Some t -> unify a t)
-	| TType ({ t_static = Some cl },params), TInst ({ cl_path = [],"Class" },[pt]) ->
-		unify (TInst (cl,params)) pt
 	| TType (t,tl) , _ ->
 		(try
 			unify (apply_params t.t_types tl t.t_type) b
@@ -648,13 +650,26 @@ let rec unify a b =
 			with
 				Not_found ->
 					if is_closed a1 then error [has_no_field a n];
-					if not (link (ref None) a f2.cf_type) then error [cannot_unify a b];
+					if not (link (ref None) a f2.cf_type) then error [];
 					a1.a_fields <- PMap.add n f2 a1.a_fields
 			) a2.a_fields;
-			if !(a1.a_status) = Opened then a1.a_status := Closed;
+			(match !(a1.a_status) with
+			| Const when not (PMap.is_empty a2.a_fields) ->
+				PMap.iter (fun n _ -> if not (PMap.mem n a2.a_fields) then error [has_extra_field a n]) a1.a_fields;
+			| Opened ->
+				a1.a_status := Closed
+			| _ -> ());
 			if !(a2.a_status) = Opened then a2.a_status := Closed;
 		with
 			Unify_error l -> error (cannot_unify a b :: l))
+	| TAnon an, TInst ({ cl_path = [],"Class" },[pt]) ->
+		(match !(an.a_status) with
+		| Statics cl -> unify (TInst (cl,List.map snd cl.cl_types)) pt
+		| _ -> error [cannot_unify a b])
+	| TAnon an, TInst ({ cl_path = [],"Enum" },[]) ->
+		(match !(an.a_status) with
+		| EnumStatics _ -> ()
+		| _ -> error [cannot_unify a b])
 	| TDynamic t , _ ->
 		if t == a then
 			()
