@@ -32,6 +32,7 @@ type context = {
 	types : (module_path, module_path) Hashtbl.t;
 	modules : (module_path , module_def) Hashtbl.t;
 	delays : (unit -> unit) list list ref;
+	constructs : (module_path , access list * type_param list * func) Hashtbl.t;
 	warn : string -> pos -> unit;
 	error : error_msg -> pos -> unit;
 	flash9 : bool;
@@ -70,7 +71,7 @@ type access_kind =
 	| AccSetField of texpr * string * t
 
 type switch_mode =
-	| CMatch of (string * (string option * t) list option)
+	| CMatch of (tenum_field * (string option * t) list option)
 	| CExpr of texpr
 
 exception Error of error_msg * pos
@@ -90,6 +91,8 @@ let unify_error_msg ctx = function
 		"Invalid type for field " ^ s ^ " :"
 	| Has_no_field (t,n) ->
 		s_type ctx t ^ " has no field " ^ n
+	| Has_extra_field (t,n) ->
+		s_type ctx t ^ " has extra field " ^ n
 	| Invalid_access (f,get,a,b) ->
 		"Inconsistent " ^ (if get then "getter" else "setter") ^ " for field " ^ f ^ " : " ^ access_str a ^ " should be " ^ access_str b
 	| Invalid_visibility n ->
@@ -134,6 +137,7 @@ let context err warn =
 	let ctx = {
 		modules = Hashtbl.create 0;
 		types = Hashtbl.create 0;
+		constructs = Hashtbl.create 0;
 		delays = ref [];
 		flash9 = Plugin.defined "flash9";
 		in_constructor = false;
@@ -252,8 +256,9 @@ let field_access ctx get f t e p =
 	| NormalAccess | F9MethodAccess ->
 		AccExpr (mk (TField (e,f.cf_name)) t p)
 	| MethodAccess m ->
-		if m = ctx.curmethod && e.eexpr = TConst TThis then
-			AccExpr (mk (TField (e,f.cf_name)) t p)
+		if m = ctx.curmethod && (match e.eexpr with TConst TThis -> true | TTypeExpr (TClassDecl c) when c == ctx.curclass -> true | _ -> false) then
+			let prefix = if ctx.flash9 && Plugin.defined "as3gen" then "$" else "" in
+			AccExpr (mk (TField (e,prefix ^ f.cf_name)) t p)
 		else if get then
 			AccExpr (mk (TCall (mk (TField (e,m)) (mk_mono()) p,[])) t p)
 		else
@@ -326,7 +331,7 @@ let rec load_normal_type ctx t p allow_no_params =
 						| Float f -> "F" ^ f
 						| _ -> assert false
 					) in
-					TEnum ({ e_path = ([],name); e_pos = p; e_doc = None; e_private = false; e_extern = true; e_types = []; e_constrs = PMap.empty },[]), true
+					TEnum ({ e_path = ([],name); e_pos = p; e_doc = None; e_private = false; e_extern = true; e_types = []; e_constrs = PMap.empty; e_names = [] },[]), true
 				| TPType t -> load_type ctx p t, false
 			) t.tparams in
 			let bparams = List.map fst tparams in
@@ -511,57 +516,6 @@ let extend_remoting ctx c t p async prot =
 		| _ -> assert false
 	)
 
-let extend_proxy ctx c t p =
-	if c.cl_super <> None then error "Cannot extend several classes" p;
-	let tclass = load_normal_type ctx t p false in
-	let tdyn = TPNormal { tpackage = []; tname = "Dynamic"; tparams = []; } in
-	let make_field f args =
-		let args = List.map (fun (name,o,t) -> name , o, Some tdyn) args in
-		let eargs = List.map (fun (name,_,_) -> EConst (Ident name) , p) args in
-		f.cf_name , (FFun (f.cf_name,None,[if f.cf_public then APublic else APrivate],[], {
-			f_args = args;
-			f_type = None;
-			f_expr = (EBlock [
-				(EReturn (Some (ECall (
-					(EConst (Ident "handleCall"),p),
-					[(ECall ((EConst (Ident "__unprotect__"),p),[EConst (String f.cf_name),p]),p); (EArrayDecl eargs,p)]
-				),p)),p)
-			],p);
-		}),p)
-	in
-	let class_fields = (match tclass with
-		| TInst (c,params) ->
-			let rec loop c =
-				PMap.fold (fun f acc ->
-					match follow f.cf_type with
-					| TFun (args,ret) ->
-						(try
-							ignore(List.assoc f.cf_name acc);
-							acc
-						with
-							Not_found -> make_field f args :: acc)
-					| _ -> acc
-				) c.cl_fields (match c.cl_super with None -> [] | Some (c,_) -> loop c)
-			in
-			List.map snd (loop c)
-		| _ ->
-			error "Proxy type parameter should be a class" p
-	) in
-	let tproxy = { tpackage = ["haxe"]; tname = "Proxy"; tparams = [TPType (TPNormal t)] } in
-	let pname = "P" ^ t.tname in
-	let class_decl = (EClass {
-		d_name = pname;
-		d_doc = None;
-		d_params = List.map (fun (s,_) -> s,[]) c.cl_types;
-		d_flags = [HExtends tproxy; HImplements t];
-		d_data = class_fields;
-	},p) in
-	let m = (!type_module_ref) ctx ("Proxy" :: t.tpackage, pname) [class_decl] p in
-	c.cl_super <- Some (match m.mtypes with
-		| [TClassDecl c2] -> (c2,List.map snd c.cl_types)
-		| _ -> assert false
-	)
-
 let extend_xml_proxy ctx c t file p =
 	let t = load_type ctx p t in
 	let file = (try Plugin.find_file file with Not_found -> file) in
@@ -602,8 +556,6 @@ let set_heritance ctx c herits p =
 			extend_remoting ctx c t p true true
 		| HExtends { tpackage = ["mt"]; tname = "AsyncProxy"; tparams = [TPType(TPNormal t)] } ->
 			extend_remoting ctx c t p true false
-		| HExtends { tpackage = ["haxe"]; tname = "Proxy"; tparams = [TPType(TPNormal t)] } when match c.cl_path with "Proxy" :: _ , _ -> false | _ -> true ->
-			extend_proxy ctx c t p
 		| HExtends t ->
 			if c.cl_super <> None then error "Cannot extend several classes" p;
 			(match t with
@@ -643,6 +595,7 @@ let type_type_params ctx path p (n,flags) =
 			e_types = [];
 			e_constrs = PMap.empty;
 			e_doc = None;
+			e_names = [];
 		} in
 		TEnum (e,[])
 	| l ->
@@ -904,9 +857,11 @@ let type_type ctx tpath p =
 			t_path = fst c.cl_path, "#" ^ snd c.cl_path;
 			t_doc = None;
 			t_pos = c.cl_pos;
-			t_type = if pub then mk_anon (PMap.map (fun f -> { f with cf_public = true }) c.cl_statics) else TAnon { a_fields = c.cl_statics; a_status = static_status };
+			t_type = TAnon {
+				a_fields = if pub then PMap.map (fun f -> { f with cf_public = true }) c.cl_statics else c.cl_statics;
+				a_status = ref (Statics c);
+			};
 			t_private = true;
-			t_static = Some c;
 			t_types = [];
 		} in
 		mk (TTypeExpr (TClassDecl c)) (TType (t_tmp,[])) p
@@ -928,9 +883,11 @@ let type_type ctx tpath p =
 			t_path = fst e.e_path, "#" ^ snd e.e_path;
 			t_doc = None;
 			t_pos = e.e_pos;
-			t_type = mk_anon fl;
+			t_type = TAnon {
+				a_fields = fl;
+				a_status = ref (EnumStatics e);
+			};
 			t_private = true;
-			t_static = None;
 			t_types = e.e_types;
 		} in
 		mk (TTypeExpr (TEnumDecl e)) (TType (t_tmp,types)) p
@@ -1087,7 +1044,7 @@ let type_matching ctx (enum,params) (e,p) ecases first_case =
 			| TEnum _ -> ()
 			| _ -> assert false
 		);
-		(name,None)
+		(c,None)
 	| ECall ((EConst (Ident name),_),el)
 	| ECall ((EConst (Type name),_),el) ->
 		let c = constr name in
@@ -1107,7 +1064,7 @@ let type_matching ctx (enum,params) (e,p) ecases first_case =
 				Some name , t
 			| _ -> invalid()
 		) el args in
-		(name,Some idents)
+		(c,Some idents)
 	| _ ->
 		invalid()
 
@@ -1561,12 +1518,15 @@ and type_switch ctx e cases def need_val p =
 				assert false
 		in
 		let constructs (el,_,e) =
-			let cl = List.map (fun c -> mk (TField (mk (TTypeExpr (TEnumDecl en)) t_dynamic p , c)) (TEnum (en,enparams)) p) el in
+			let cl = List.map (fun c -> mk (TField (mk (TTypeExpr (TEnumDecl en)) t_dynamic p , c.ef_name)) (TEnum (en,enparams)) p) el in
 			(cl,e)
 		in
+		let indexes (el,vars,e) =
+			List.map (fun c -> c.ef_index) el, vars, e
+		in
 		let cases = List.map matchs cases in
 		match !has_params with
-		| true -> mk (TMatch (e,(en,enparams),cases,def)) t p
+		| true -> mk (TMatch (e,(en,enparams),List.map indexes cases,def)) t p
 		| false -> mk (TSwitch (e,List.map constructs cases,def)) t p
 
 and type_access ctx e p get =
@@ -1708,7 +1668,9 @@ and type_expr ctx ?(need_val=true) (e,p) =
 			((f,e) :: l, PMap.add f cf acc)
 		in
 		let fields , types = List.fold_left loop ([],PMap.empty) fl in
-		mk (TObjectDecl fields) (mk_anon types) p
+		let x = ref Const in
+		ctx.opened <- x :: ctx.opened;
+		mk (TObjectDecl (List.rev fields)) (TAnon { a_fields = types; a_status = x }) p
 	| EArrayDecl el ->
 		let t = ref (mk_mono()) in
 		let el = List.map (fun e ->
@@ -2315,17 +2277,18 @@ let init_class ctx c p herits fields =
 					type_type_params ctx ([],name) p (n,[])
 				| _ -> error "This notation is not allowed because it can't be checked" p
 			) params in
+			let stat = List.mem AStatic access in
 			let ctx = { ctx with
 				curclass = c;
 				curmethod = name;
 				tthis = tthis;
-				type_params = params @ ctx.type_params;
+				type_params = if stat then params  else params @ ctx.type_params;
 			} in
 			let ret = type_opt ctx p f.f_type in
 			let args = List.map (fun (name,opt,t) -> name , opt, type_opt ~param:opt ctx p t) f.f_args in
 			let t = TFun (args,ret) in
-			let stat = List.mem AStatic access in
 			let constr = (name = "new") in
+			if constr && c.cl_interface then error "An interface cannot have a constructor" p;
 			if c.cl_interface && not stat && (match f.f_expr with EBlock [] , _ -> false | _ -> true) then error "An interface method cannot have a body" p;
 			if constr then (match f.f_type with
 				| None | Some (TPNormal { tpackage = []; tname = "Void" }) -> ()
@@ -2426,33 +2389,31 @@ let init_class ctx c p herits fields =
 	) fields in
 	c.cl_ordered_statics <- List.rev c.cl_ordered_statics;
 	c.cl_ordered_fields <- List.rev c.cl_ordered_fields;
-	(* define an default inherited constructor *)
-	(match c.cl_constructor, c.cl_super with
-	| None , Some ({ cl_constructor = Some f; cl_types = tl } as csuper, cparams) ->
-		let t = apply_params tl cparams (field_type f) in
-		(match follow t with
-		| TFun (args,r) ->
-			let n = ref 0 in
-			let args = List.map (fun (_,b,t) -> incr n; "p" ^ string_of_int (!n) , b, t) args in
-			let eargs = List.map (fun (n,_,t) -> mk (TLocal n) t p) args in
-			let func = {
-				tf_args = args;
-				tf_type = t_void ctx;
-				tf_expr = mk (TCall (mk (TConst TSuper) (TInst (csuper,cparams)) p,eargs)) r p;
-			} in
-			c.cl_constructor <- Some {
-				cf_name = "new";
-				cf_type = t;
-				cf_get = NormalAccess;
-				cf_set = NoAccess;
-				cf_doc = None;
-				cf_expr = Some (mk (TFunction func) t p);
-				cf_public = f.cf_public;
-				cf_params = f.cf_params;
-			}
-		| _ -> assert false)
-	| _ , _ ->
-		());
+	(* 
+		define a default inherited constructor.
+		This is actually pretty tricky since we can't assume that the constructor of the
+		superclass has been defined yet because type structure is not stabilized wrt recursion.		
+	*)
+	let rec define_constructor ctx c =
+		try
+			Some (Hashtbl.find ctx.constructs c.cl_path)
+		with Not_found ->
+			match c.cl_super with
+			| None -> None
+			| Some (csuper,_) ->
+				match define_constructor ctx csuper with
+				| None -> None
+				| Some (acc,pl,f) as infos ->
+					let p = c.cl_pos in
+					let esuper = (ECall ((EConst (Ident "super"),p),List.map (fun (n,_,_) -> (EConst (Ident n),p)) f.f_args),p) in
+					let acc = (if csuper.cl_extern && acc = [] then [APublic] else acc) in
+					let _, _, cf, delayed = loop_cf (FFun ("new",None,acc,pl,{ f with f_expr = esuper })) p in
+					c.cl_constructor <- Some cf;
+					Hashtbl.add ctx.constructs c.cl_path (acc,pl,f);
+					ctx.delays := [delayed] :: !(ctx.delays);
+					infos
+	in
+	ignore(define_constructor ctx c);
 	fl
 
 let type_module ctx m tdecls loadp =
@@ -2479,6 +2440,12 @@ let type_module ctx m tdecls loadp =
 			let priv = List.mem HPrivate d.d_flags in
 			let path = decl_with_name d.d_name p priv in
 			let c = mk_class path p d.d_doc priv in
+			(* store the constructor for later usage *)
+			List.iter (fun (cf,_) ->
+				match cf with
+				| FFun ("new",_,acc,pl,f) -> Hashtbl.add ctx.constructs path (acc,pl,f)
+				| _ -> ()
+			) d.d_data;
 			decls := TClassDecl c :: !decls
 		| EEnum d ->
 			let priv = List.mem EPrivate d.d_flags in
@@ -2491,6 +2458,7 @@ let type_module ctx m tdecls loadp =
 				e_private = priv;
 				e_extern = List.mem EExtern d.d_flags || d.d_data = [];
 				e_constrs = PMap.empty;
+				e_names = [];
 			} in
 			decls := TEnumDecl e :: !decls
 		| ETypedef d ->
@@ -2501,8 +2469,7 @@ let type_module ctx m tdecls loadp =
 				t_pos = p;
 				t_doc = d.d_doc;
 				t_private = priv;
-				t_types = [];
-				t_static = None;
+				t_types = [];				
 				t_type = mk_mono();
 			} in
 			decls := TTypeDecl t :: !decls
@@ -2517,6 +2484,7 @@ let type_module ctx m tdecls loadp =
 	let ctx = {
 		modules = ctx.modules;
 		delays = ctx.delays;
+		constructs = ctx.constructs;
 		types = ctx.types;
 		warn = ctx.warn;
 		error = ctx.error;
@@ -2592,15 +2560,26 @@ let type_module ctx m tdecls loadp =
 			let e = get_enum d.d_name in
 			ctx.type_params <- e.e_types;
 			let et = TEnum (e,List.map snd e.e_types) in
+			let names = ref [] in
+			let index = ref 0 in
 			List.iter (fun (c,doc,t,p) ->
 				if c = "name" && Plugin.defined "js" then error "This identifier cannot be used in Javascript" p;
 				let t = (match t with
 					| [] -> et
-					| l -> TFun (List.map (fun (s,b,t) -> s, b, load_type ctx p t) l, et)
+					| l -> TFun (List.map (fun (s,opt,t) -> s, opt, load_type_opt ~param:opt ctx p (Some t)) l, et)
 				) in
-				if PMap.mem c e.e_constrs then error ("Duplicate constructor " ^ c) p;
-				e.e_constrs <- PMap.add c { ef_name = c; ef_type = t; ef_pos = p; ef_doc = doc } e.e_constrs
-			) d.d_data
+				if PMap.mem c e.e_constrs then error ("Duplicate constructor " ^ c) p;				
+				e.e_constrs <- PMap.add c {
+					ef_name = c;
+					ef_type = t;
+					ef_pos = p;
+					ef_doc = doc;
+					ef_index = !index;
+				} e.e_constrs;
+				incr index;
+				names := c :: !names;
+			) d.d_data;
+			e.e_names <- List.rev !names;
 		| ETypedef d ->
 			let t = get_tdef d.d_name in
 			ctx.type_params <- t.t_types;
